(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('react')) :
    typeof define === 'function' && define.amd ? define(['exports', 'react'], factory) :
    (global = global || self, factory(global.reto = {}, global.React));
}(this, (function (exports, React) { 'use strict';

    var React__default = 'default' in React ? React['default'] : React;

    class NoStoreError {
        constructor(message, store) {
            this.message = message;
            this.store = store;
        }
    }

    class Container {
        constructor(state) {
            this.state = state;
            this.subscribers = new Set();
            this.initialized = false;
        }
        notify() {
            for (const subscriber of this.subscribers) {
                subscriber();
            }
        }
    }

    const defaultStoreValue = Symbol();
    function getStoreContext(s) {
        if (!s.Context) {
            s.Context = React__default.createContext(new Container(defaultStoreValue));
        }
        return s.Context;
    }

    function Consumer(props) {
        const store = useStore(props.of);
        return props.children(store);
    }
    function useStore(s, deps) {
        const name = s.displayName || s.name;
        React.useDebugValue(name);
        const Context = getStoreContext(s);
        const container = React.useContext(Context);
        const [state, setState] = React.useState(container.state);
        const depsRef = React.useRef([]);
        React.useEffect(() => {
            const subscriber = () => {
                if (!deps) {
                    setState(container.state);
                }
                else {
                    const oldDeps = depsRef.current;
                    const newDeps = deps(container.state);
                    if (compare(oldDeps, newDeps)) {
                        setState(container.state);
                    }
                    depsRef.current = newDeps;
                }
            };
            container.subscribers.add(subscriber);
            return () => {
                container.subscribers.delete(subscriber);
            };
        }, []);
        if (state === defaultStoreValue) {
            throw new NoStoreError(`No store context of "${name}" found. And "${name}" doesn't have defaultValue. Either render a Provider or set a defaultValue for this Store.`, s);
        }
        return state;
    }
    function useStoreOptionally(s, deps) {
        try {
            const store = useStore(s, deps);
            return [store, true];
        }
        catch (e) {
            if (e instanceof NoStoreError) {
                return [undefined, false];
            }
            else {
                throw e;
            }
        }
    }
    function compare(oldDeps, newDeps) {
        if (oldDeps.length !== newDeps.length) {
            return true;
        }
        for (const index in newDeps) {
            if (oldDeps[index] !== newDeps[index]) {
                return true;
            }
        }
        return false;
    }

    const Executor = React.memo(function Executor(props) {
        var _a;
        const args = (_a = props.args) !== null && _a !== void 0 ? _a : [];
        const result = props.useStore(...args);
        props.container.state = result;
        props.container.initialized = true;
        React.useEffect(() => {
            props.onChange(result);
            props.container.notify();
        });
        return null;
    }, (prevProps, nextProps) => {
        if (!nextProps.memo) {
            return false;
        }
        else {
            return argsAreEqual(prevProps.args, nextProps.args);
        }
    });
    function argsAreEqual(args1, args2) {
        if (args1 === args2)
            return true;
        if (args1.length !== args2.length)
            return false;
        return args1.every((element, index) => element === args2[index]);
    }

    const Provider = function (props) {
        const Context = getStoreContext(props.of);
        const containerRef = React.useRef();
        if (!containerRef.current) {
            containerRef.current = new Container();
        }
        const container = containerRef.current;
        const checkRef = React.useRef();
        function onChange(value) {
            if (props.storeRef) {
                props.storeRef.current = value;
            }
            checkRef.current.onInitialize();
        }
        return (React.createElement(React.Fragment, null,
            React.createElement(Executor, { useStore: props.of, args: props.args, onChange: onChange, memo: props.memo, container: container }),
            React.createElement(Context.Provider, { value: container },
                React.createElement(Checker, { container: container, ref: checkRef }, props.children))));
    };
    Provider.displayName = 'Provider';
    Provider.defaultProps = {
        args: [],
    };
    const Checker = React.forwardRef((props, ref) => {
        const [initialized, setInitialized] = React.useState(props.container.initialized);
        React.useImperativeHandle(ref, () => ({
            onInitialize: () => {
                setInitialized(true);
            }
        }));
        return initialized && (React.createElement(React.Fragment, null, props.children));
    });
    Checker.displayName = 'Checker';

    exports.Consumer = Consumer;
    exports.Provider = Provider;
    exports.useStore = useStore;
    exports.useStoreOptionally = useStoreOptionally;

    Object.defineProperty(exports, '__esModule', { value: true });

})));
